import java.util.Vector;

public class TransformationRule
{ String name;
  boolean lazy;
  boolean primary;
  Attribute source;
  Expression guard;
  Vector targets = new Vector(); // of Attribute
  Statement body;

  public TransformationRule(String n, boolean l, boolean p)
  { name = n; lazy = l; primary = p; }

  public void setGuard(Expression e)
  { guard = e; }

  public void setSource(Attribute s)
  { source = s; }

  public void setBody(Statement s)
  { body = s; }

  public void addTarget(Attribute trg)
  { targets.add(trg); }

  public void addTargets(Vector trgs)
  { targets.addAll(trgs); }

  public Attribute getSource()
  { return source; }

  public String toString()
  { String res = "";
    if (lazy)
    { res = res + "@lazy\n"; }
    if (primary)
    { res = res + "@primary\n"; }
    res = "rule " + name + "\n" +
       "  transform " + source.getName() + " : " + source.getType() + "\n" +
       "  to ";
    for (int i = 0; i < targets.size(); i++)
    { Attribute trg = (Attribute) targets.get(i);
      res = res + trg.getName() + " : " + trg.getType();
      if (i < targets.size() - 1)
      { res = res + ", "; }
      res = res + "\n    ";
    }
    res = res + "  { guard: " + guard + "\n  " +
         body + "  }\n";
    return res;
  }

  public boolean isPrimary()
  { return primary; }

  public boolean isLazy()
  { return lazy; }

  public boolean hasOutputType(Entity trg)
  { boolean res = false;
    for (int i = 0; i < targets.size(); i++)
    { Attribute att = (Attribute) targets.get(i);
      Type t = att.getType();
      if (t.isEntity())
      { Entity et = t.getEntity(); 
        if (et == trg || Entity.isAncestor(trg,et))
        { return true; }
      }
    }
    return res;
  }

  public Constraint toConstraint(Vector entities, Vector types)
  { // guard => T1->exists( t1 | ... Tn->exists( tn | AddToTrace & self.op(t1,...,tn) ) ... )
    String srcname = source.getName();
    Type srct = source.getType();
    Entity srcent = srct.getEntity();

    String opcall = "self." + name + "op(";
   
    Entity trent = (Entity) ModelElement.lookupByName("$Trace", entities);
    Entity outent = (Entity) ModelElement.lookupByName("$OUT", entities);

    Type trentt = new Type(trent);
    Type outentt = new Type(outent);

    Expression trentexp = new BasicExpression("$Trace");
    trentexp.setUmlKind(Expression.CLASSID);
    // trentexp.setType(new Type("Set", null));
    trentexp.setElementType(trentt);
    // trentexp.setEntity(tent);
    Expression tracevar = new BasicExpression("$trace");
    tracevar.setUmlKind(Expression.ROLE);
    tracevar.setElementType(trentt);
    tracevar.setType(new Type("Sequence", null));
    tracevar.setMultiplicity(ModelElement.MANY);

    // trent, outent needed
    // trent = (Entity) ModelElement.lookupByName("$Trace", entities);
    // outent = (Entity) ModelElement.lookupByName("$OUT", entities);

    Vector opparams = new Vector();

    Expression add2targets = new BasicExpression("true");

    for (int i = 0; i < targets.size(); i++)
    { Attribute trg = (Attribute) targets.get(i);
      String tname = trg.getName();
      opcall = opcall + tname;
      if (i < targets.size() - 1)
      { opcall = opcall + ","; }
      Expression tvar = new BasicExpression(tname);
      tvar.setUmlKind(Expression.VARIABLE);
      tvar.setType(trg.getType());
      tvar.setElementType(trg.getType());

      opparams.add(tvar);
      Expression trvar = new BasicExpression("$" + tname);
      trvar.setUmlKind(Expression.VARIABLE);
      trvar.setType(trentt);
      trvar.setElementType(trentt);
      Expression intent = new BinaryExpression(":", trvar, trentexp);
      Expression trintrace = new BinaryExpression(":", trvar, tracevar);
      BasicExpression targetf = new BasicExpression("target");
      targetf.setUmlKind(Expression.ROLE);
      targetf.setMultiplicity(ModelElement.ONE);
      targetf.setElementType(outentt);
      targetf.setType(outentt);
      targetf.setObjectRef(trvar);
      Expression targeteq = new BinaryExpression("=", targetf, tvar);
      BinaryExpression conj = new BinaryExpression("&", trintrace, targeteq);
      Expression add2target = new BinaryExpression("#", intent, conj);
      add2target.setBrackets(true);
      add2targets = Expression.simplifyAnd(add2targets,add2target);
    }
    opcall = opcall + ")";
    BasicExpression invoke = new BasicExpression(opcall,0); 
    Expression opcallf = invoke.checkIfSetExpression();
    opcallf.setUmlKind(Expression.UPDATEOP); 
    add2targets = Expression.simplifyAnd(add2targets, opcallf);

    for (int j = targets.size() - 1; j >= 0; j--)
    { Attribute trg = (Attribute) targets.get(j);
      String tname = trg.getName();  
      Type ttype = trg.getType();
      Expression tvar = (Expression) opparams.get(j);
      Expression texp = new BasicExpression(ttype + "");
      texp.setUmlKind(Expression.CLASSID);
      texp.setType(new Type("Set", null));
      texp.setElementType(ttype);
      Expression intent = new BinaryExpression(":", tvar, texp);
      add2targets = new BinaryExpression("#", intent, add2targets);
    }

    Expression selfexp = new BasicExpression("self");
    selfexp.setUmlKind(Expression.VARIABLE);
    selfexp.setEntity(srcent);
    selfexp.setType(srct);
    selfexp.setElementType(srct);

    Statement opcode = body.substituteEq(srcname, selfexp);
    opcode.setEntity(srcent); 

    BehaviouralFeature bf = new BehaviouralFeature(name + "op");
    bf.setParameters(targets);
    bf.setEntity(srcent);
    srcent.addOperation(bf);
    Vector env = new Vector(); 
    env.addAll(targets); 
    opcode.typeCheck(types,entities,env); 
    bf.setActivity(opcode); 
    
    Expression newguard = guard.substituteEq(srcname, selfexp);

    Constraint con = new Constraint(newguard,add2targets);
    con.setOwner(srcent);  
    return con; 
  } 
}

